/*

    File: ext2_dir.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2fs.h"
#include "types.h"
#include "common.h"
#include "intrface.h"
#include "hdaccess.h"
#include "dir.h"
static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

#define DIRENT_DELETED_FILE	4
/*
 * list directory
 */

#define LONG_OPT	0x0001
typedef struct ext_dir ext2_file_data;
struct ext_dir
{
  char	name[EXT2_NAME_LEN];
  time_t modtime;
  __u16  mode;
  __u16  uid;
  __u16  gid;
  __u8   attr;
  __u32  start;		/* starting cluster number */
  __u64	 size;			/* file size (in bytes) */
};

struct list_dir_struct {
	t_file_data *dir_list;
	t_file_data *current_file;
	ext2_filsys current_fs;
	int flags;
};

static char ftypelet (int bits);
void mode_string (const int mode, char *str);
static void ext2_write_display(WINDOW *window, const void*private_data);
static void ext2_write_rapport(const void*private_data);
static int ext2_is_directory(const void*private_data);
static long int ext2_get_inode(const void*private_data);

static char ftypelet (int bits)
{
#ifdef LINUX_S_ISBLK
  if (LINUX_S_ISBLK (bits))
    return 'b';
#endif
  if (LINUX_S_ISCHR (bits))
    return 'c';
  if (LINUX_S_ISDIR (bits))
    return 'd';
  if (LINUX_S_ISREG (bits))
    return '-';
#ifdef LINUX_S_ISFIFO
  if (LINUX_S_ISFIFO (bits))
    return 'p';
#endif
#ifdef LINUX_S_ISLNK
  if (LINUX_S_ISLNK (bits))
    return 'l';
#endif
#ifdef LINUX_S_ISSOCK
  if (LINUX_S_ISSOCK (bits))
    return 's';
#endif
#ifdef LINUX_S_ISMPC
  if (LINUX_S_ISMPC (bits))
    return 'm';
#endif
#ifdef LINUX_S_ISNWK
  if (LINUX_S_ISNWK (bits))
    return 'n';
#endif
#ifdef LINUX_S_ISDOOR
  if (LINUX_S_ISDOOR (bits))
    return 'D';
#endif
#ifdef LINUX_S_ISCTG
  if (LINUX_S_ISCTG (bits))
    return 'C';
#endif
#ifdef LINUX_S_ISOFD
  if (LINUX_S_ISOFD (bits))
    /* off line, with data  */
    return 'M';
#endif
#ifdef LINUX_S_ISOFL
  /* off line, with no data  */
  if (LINUX_S_ISOFL (bits))
    return 'M';
#endif
  return '?';
}

void mode_string (const int mode, char *str)
{
  str[0] = ftypelet(mode);
  str[1] = mode & LINUX_S_IRUSR ? 'r' : '-';
  str[2] = mode & LINUX_S_IWUSR ? 'w' : '-';
  str[3] = mode & LINUX_S_IXUSR ? 'x' : '-';
  str[4] = mode & LINUX_S_IRGRP ? 'r' : '-';
  str[5] = mode & LINUX_S_IWGRP ? 'w' : '-';
  str[6] = mode & LINUX_S_IXGRP ? 'x' : '-';
  str[7] = mode & LINUX_S_IROTH ? 'r' : '-';
  str[8] = mode & LINUX_S_IWOTH ? 'w' : '-';
  str[9] = mode & LINUX_S_IXOTH ? 'x' : '-';
  str[10]='\0';
#ifdef LINUX_S_ISUID
  if (mode & LINUX_S_ISUID)
  {
    if (str[3] != 'x')
      /* Set-uid, but not executable by owner.  */
      str[3] = 'S';
    else
      str[3] = 's';
  }
#endif
#ifdef LINUX_S_ISGID
  if (mode & LINUX_S_ISGID)
  {
    if (str[6] != 'x')
      /* Set-gid, but not executable by group.  */
      str[6] = 'S';
    else
      str[6] = 's';
  }
#endif
#ifdef LINUX_S_ISVTX
  if (mode & LINUX_S_ISVTX)
  {
    if (str[9] != 'x')
      /* Sticky, but not executable by others.  */
      str[9] = 'T';
    else
      str[9] = 't';
  }
#endif
}

/*
 * I/O Manager routine prototypes
 */
static errcode_t my_open(const char *dev, int flags, io_channel *channel);
static errcode_t my_close(io_channel channel);
static errcode_t my_set_blksize(io_channel channel, int blksize);
static errcode_t my_read_blk(io_channel channel, unsigned long block, int count, void *buf);
static errcode_t my_write_blk(io_channel channel, unsigned long block, int count, const void *buf);
static errcode_t my_flush(io_channel channel);
static errcode_t my_write_byte(io_channel channel, unsigned long offset, int size, const void *buf);


static io_channel alloc_io_channel(t_param_disk disk_car,t_diskext partition);

static struct struct_io_manager my_struct_manager = {
        EXT2_ET_MAGIC_IO_MANAGER,
        "TestDisk I/O Manager",
        my_open,
        my_close,
        my_set_blksize,
        my_read_blk,
        my_write_blk,
        my_flush,
	my_write_byte
};

io_manager my_io_manager = &my_struct_manager;
/*
 * Macro taken from unix_io.c
 * For checking structure magic numbers...
 */

#define EXT2_CHECK_MAGIC(struct, code) \
          if ((struct)->magic != (code)) return (code)

/*
 * Allocate libext2fs structures associated with I/O manager
 */
static io_channel alloc_io_channel(t_param_disk disk_car,t_diskext partition)
{
  io_channel     ioch;
  t_my_data	*my_data;
#ifdef DEBUG
  ecrit_rapport("alloc_io_channel start\n");
#endif
  ioch = (io_channel)malloc(sizeof(struct struct_io_channel));
  if (!ioch)
	  return NULL;
  memset(ioch, 0, sizeof(struct struct_io_channel));
  ioch->magic = EXT2_ET_MAGIC_IO_CHANNEL;
  ioch->manager = my_io_manager;
  ioch->name = (char *)malloc(strlen(partition->name)+1);
  if (!ioch->name) {
	  free(ioch);
	  return NULL;
  }
  strcpy(ioch->name, partition->name);
  my_data=MALLOC(sizeof(*my_data));
  my_data->partition=partition;
  my_data->disk_car=disk_car;
  ioch->private_data = my_data;
  ioch->block_size = 1024; /* The smallest ext2fs block size */
  ioch->read_error = 0;
  ioch->write_error = 0;
#ifdef DEBUG
  ecrit_rapport("alloc_io_channel end\n");
#endif
  return ioch;
}

static errcode_t my_open(const char *dev, int flags, io_channel *channel)
{
  *channel = (io_channel*)dev;
#ifdef DEBUG
  ecrit_rapport("my_open done\n");
#endif
  return 0;
};

static errcode_t my_close(io_channel channel)
{
  FREE(channel->private_data);
  FREE(channel->name);
  FREE(channel);
#ifdef DEBUG
  ecrit_rapport("my_close done\n");
#endif
  return 0;
}

static errcode_t my_set_blksize(io_channel channel, int blksize)
{
  channel->block_size = blksize;
#ifdef DEBUG
  ecrit_rapport("my_set_blksize done\n");
#endif
  return 0;
}

static errcode_t my_read_blk(io_channel channel, unsigned long block, int count, void *buf)
{
  size_t size;
  t_my_data *my_data=(t_my_data*)channel->private_data;
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);

  size = (count < 0) ? -count : count * channel->block_size;
#ifdef DEBUG
  ecrit_rapport("my_read_blk start\n");
  ecrit_rapport("size=%u, block=%u name=%s\n",size,block,my_data->partition->name);
#endif
  if(hd_read_LBA(my_data->disk_car,size/512,buf,my_data->partition->lba+(block*channel->block_size/512))!=0)
    return 1;
#ifdef DEBUG
  ecrit_rapport("my_read_blk done\n");
#endif
  return 0;
}

static errcode_t my_write_blk(io_channel channel, unsigned long block, int count, const void *buf)
{
  t_my_data *my_data=(t_my_data*)channel;
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
//  if(hd_write_LBA(my_data->disk_car,count*channel->block_size/512,buf,my_data->partition->lba+(block*channel->block_size/512))!=0)
    return 1;
//  return 0;
}

static errcode_t my_flush(io_channel channel)
{
  return 0;
}

static errcode_t my_write_byte(io_channel channel, unsigned long offset, int size, const void *buf)
{
  return 1;
}

static int list_dir_proc(struct ext2_dir_entry *dirent,
			 int	offset,
			 int	blocksize,
			 char	*buf,
			 void	*private)
{
  struct ext2_inode	inode;
  ext2_ino_t		ino;
  char			name[EXT2_NAME_LEN];
  int			thislen;
  struct list_dir_struct *ls = (struct list_dir_struct *) private;
  t_file_data *new_file=MALLOC(sizeof(*new_file));
  ext2_file_data *new_ext_element=MALLOC(sizeof(*new_ext_element));

  thislen = ((dirent->name_len & 0xFF) < EXT2_NAME_LEN) ?
    (dirent->name_len & 0xFF) : EXT2_NAME_LEN;
  strncpy(name, dirent->name, thislen);
  name[thislen] = '\0';
  ino = dirent->inode;
  new_file->prev=ls->current_file;
  new_file->next=NULL;
  new_file->file_data=new_ext_element;
  if (ino) {
    if (ext2fs_read_inode(ls->current_fs,ino, &inode))
    {
      FREE(new_ext_element);
      FREE(new_file);
      return 0;
    }
    new_ext_element->modtime=inode.i_mtime;
  } else {
    memset(&inode, 0, sizeof(struct ext2_inode));
    new_ext_element->modtime=0;
  }
  new_ext_element->mode=inode.i_mode;
  new_ext_element->uid=inode.i_uid;
  new_ext_element->gid=inode.i_gid;
  new_ext_element->start=ino;
  new_ext_element->size=LINUX_S_ISDIR(inode.i_mode)?inode.i_size:
    inode.i_size| ((__u64)inode.i_size_high << 32);
  memcpy(new_ext_element->name,name,EXT2_NAME_LEN);
  if(ls->current_file)
    ls->current_file->next=new_file;
  else
    ls->dir_list=new_file;
  ls->current_file=new_file;
  return 0;
}

static int ext2_is_directory(const void*private_data)
{
  const ext2_file_data *file=(const ext2_file_data*)private_data;
  return LINUX_S_ISDIR(file->mode);
}

static long int ext2_get_inode(const void*private_data)
{
  const ext2_file_data *file=(const ext2_file_data*)private_data;
  return file->start;
}

static void ext2_write_rapport(const void*private_data)
{
  const ext2_file_data *file=(const ext2_file_data*)private_data;
  struct tm		*tm_p;
  char		datestr[80];
  char str[10];
  if(file->modtime)
  {
    tm_p = localtime(&file->modtime);
    sprintf(datestr, "%2d-%s-%4d %02d:%02d",
	tm_p->tm_mday, monstr[tm_p->tm_mon],
	1900 + tm_p->tm_year, tm_p->tm_hour,
	tm_p->tm_min);
  } else {
    strcpy(datestr, "                 ");
  }
  mode_string(file->mode,str);
  ecrit_rapport("%s %5d  %5d   ", 
      str, file->uid, file->gid);
  ecrit_rapport("%7lld", file->size);
  ecrit_rapport(" %s %s\n", datestr, file->name);
}


static void ext2_write_display(WINDOW *window, const void*private_data)
{
  const ext2_file_data *file=(const ext2_file_data*)private_data;
  struct tm		*tm_p;
  char str[11];
  char		datestr[80];
  if(file->modtime!=0)
  {
    tm_p = localtime(&file->modtime);
    sprintf(datestr, "%2d-%s-%4d %02d:%02d",
	tm_p->tm_mday, monstr[tm_p->tm_mon],
	1900 + tm_p->tm_year, tm_p->tm_hour,
	tm_p->tm_min);
  } else {
    strcpy(datestr, "                 ");
  }
  mode_string(file->mode,str);
  wdoprintf(window, "%s %5d %5d   ", 
      str, file->uid, file->gid);
  wdoprintf(window, "%7lld", file->size);
  wdoprintf(window, " %s %s\n", datestr, file->name);
}

t_file_data *ext2_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster);
t_file_data *ext2_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster)
{
  struct list_dir_struct *ls=(struct list_dir_struct*)dir_data->private_dir_data;
  ls->dir_list=NULL;
  ls->current_file=NULL;
  ext2fs_dir_iterate(ls->current_fs, cluster, ls->flags, 0, list_dir_proc, ls);
  return ls->dir_list;
}

int dir_partition_ext2(WINDOW *window,t_param_disk disk_car,t_diskext partition, const int debug)
{
  errcode_t       retval;
  struct list_dir_struct ls;
  io_channel ioch;
  ls.dir_list=NULL;
  ls.current_file=NULL;
  aff_buffer(BUFFER_RESET,"Q");
  wmove(window,4,0);
  aff_part(window,1,disk_car,partition);
  aff_part_rapport(1,disk_car,partition);
  ecrit_rapport("\n");
  ioch=alloc_io_channel(disk_car,partition);
  retval = ext2fs_open (ioch, 0, partition->boot_sector, partition->blocksize, my_io_manager, &ls.current_fs);
  if(retval)
  {
    aff_buffer(BUFFER_ADD,"Can't open ext2fs (retval=%u)\n",retval);
    aff_buffer(BUFFER_DISPLAY,"Q",window);
    ext2fs_close (ls.current_fs);
    return -1;
  }
  ls.flags = 0; //DIRENT_FLAG_INCLUDE_EMPTY;

  {
    t_dir_data dir_data;
    dir_data.window=window;
    dir_data.write_rapport=ext2_write_rapport;
    dir_data.write_display=ext2_write_display;
    dir_data.is_directory=ext2_is_directory;
    dir_data.get_cluster_inode=ext2_get_inode;
    dir_data.debug=debug;
    dir_data.private_dir_data=&ls;
    dir_data.get_dir=ext2_dir;
//    dir_aff(disk_car,partition,&dir_data,ls.dir_list,EXT2_ROOT_INO);
    dir_partition(disk_car,partition,&dir_data,EXT2_ROOT_INO);
  }
  ext2fs_close (ls.current_fs);
  return 0;
}

